import base64
from os import environ, path

import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives.asymmetric import padding

from .exceptions import TokenError


class PasswordManager(object):

    def __init__(self, endpoint=environ.get('GSENHA_ENDPOINT'),
                 user=environ.get('GSENHA_USER'), password=environ.get('GSENHA_PASS'),
                 key=environ.get('GSENHA_KEY', environ.get('GSENHA_KEY_PATH')), verify=None):
        self._user = user
        self._password = password
        self._verify = verify
        self._rsa_verifier = self._load_key(key)
        self._headers = {
            'Content-Type': 'application/json'
        }
        self._gsenha_endpoint = endpoint
        self._token = self._get_token()

    def _load_key(self, gsenha_key):
        if gsenha_key:
            gsenha_key = self.__load_key_from_file(gsenha_key)

        try:
            return load_pem_private_key(
                data=gsenha_key.encode('ascii'),
                password=None,
                backend=default_backend()
            )
        except ValueError:
            raise Exception('Error loading private key')

    def __load_key_from_file(self, key_path):
        if path.exists(key_path):
            with open(key_path) as opened_key:
                return opened_key.read()
        else:
            raise ValueError('Key File does not exist')

    def _get_token(self):
        try:
            token_response = requests.post(
                '{}/login'.format(self._gsenha_endpoint),
                headers=self._headers,
                json={'username': self._user, 'password': self._password},
                verify=self._verify
            )
        except Exception as e:
            raise TokenError('There was error requesting token. Error was: {}'.format(str(e)))
        else:
            if token_response.ok:
                token_json = token_response.json()
                if token_json:
                    return token_json.get('token')
            else:
                raise TokenError(
                    'Response status code to get token is not ok. Status code was: {}'.format(
                        token_response.status_code
                    )
                )

    def _get_password(self, folder, name):
        headers = self._headers.copy()
        headers['Authorization'] = 'Bearer {}'.format(self._token)

        response = requests.post(
            '{}/search/password'.format(self._gsenha_endpoint),
            headers=headers,
            json={'folder': folder, 'name': name},
            verify=self._verify
        )

        password_json = response.json()
        if password_json.get('status') == 'success':
            return password_json.get('password')

    def _decrypt(self, value):
        raw_cipher_data = base64.b64decode(value)
        return str(self._rsa_verifier.decrypt(raw_cipher_data, padding.PKCS1v15()).decode('utf-8'))

    def _decode_b64(self, a_str):
        # If string % 4 -add back '='
        l_mod = len(a_str) % 4
        if l_mod:
            a_str += '=' * (4 - l_mod)
        return (base64.urlsafe_b64decode(a_str)).decode('utf-8')

    def get_passwords(self, folder, *names):
        return_passwords = dict()
        for name in names:
            password = self._get_password(folder, name)
            if password:
                if password.get('vault'):
                    return_passwords[name] = {
                        'url': self._decode_b64(password['url']),
                        'login': self._decode_b64(password['login']),
                        'password': self._decode_b64(password['passwd']),
                        'description': self._decode_b64(password['description']),
                    }
                else:
                    return_passwords[name] = {
                        'url': self._decrypt(password['url']),
                        'login': self._decrypt(password['login']),
                        'password': self._decrypt(password['passwd']),
                        'description': self._decrypt(password['description']),
                    }
        return return_passwords



import base64
from unittest import TestCase
try:
    from unittest.mock import patch, Mock
except ImportError:
    from mock import patch, Mock

from cryptography.hazmat.primitives.asymmetric import padding

from gsenha import PasswordManager
from gsenha.exceptions import TokenError


class PasswordManagerTest(TestCase):

    def test_invalid_key_raises_exception(self):
        with self.assertRaises(Exception) as context_manager:
            PasswordManager(key='')

        self.assertEqual(str(context_manager.exception), 'Error loading private key')

    def test_invalid_key_path_raises_exception(self):
        with self.assertRaises(ValueError) as context_manager:
            PasswordManager(key='/this/path/does/not/exists')

        self.assertEqual(str(context_manager.exception), 'Key File does not exist')

    @patch('requests.post')
    def test_response_to_get_token_is_not_ok(self, mock_post):
        mock_response = Mock()
        mock_response.ok = False
        mock_response.status_code = 500
        mock_post.return_value = mock_response

        with self.assertRaises(TokenError) as context_manager:
            PasswordManager(key='tests/fixtures/privkey.pem')

        self.assertEqual(
            str(context_manager.exception),
            'Response status code to get token is not ok. Status code was: 500'
        )

    @patch('requests.post')
    def test_response_to_get_token_returns_error(self, mock_post):
        mock_post.side_effect = Exception('Boom!')

        with self.assertRaises(Exception) as context_manager:
            PasswordManager(key='tests/fixtures/privkey.pem')

        self.assertEqual(
            str(context_manager.exception),
            'There was error requesting token. Error was: Boom!'
        )

    @patch('requests.post')
    def test_response_to_get_token_is_ok_but_there_is_no_token(self, mock_post):
        mock_response = Mock()
        mock_response.ok = True
        mock_response.json.return_value = {}
        mock_post.return_value = mock_response

        pm = PasswordManager(key='tests/fixtures/privkey.pem')

        self.assertIsNone(pm._token)

    @patch('requests.post')
    def test_response_to_get_token_is_ok(self, mock_post):
        mock_response = Mock()
        mock_response.ok = True
        mock_response.json.return_value = {'token': 'my-fake-token'}
        mock_post.return_value = mock_response

        pm = PasswordManager(key='tests/fixtures/privkey.pem')

        self.assertEqual(pm._token, 'my-fake-token')

    @patch.object(PasswordManager, '_get_token', new=lambda x: 'my-fake-token')
    def test_try_to_get_password_but_password_name_is_not_passed(self):
        pm = PasswordManager(key='tests/fixtures/privkey.pem')

        passwords = pm.get_passwords('folder-name')

        self.assertDictEqual(passwords, {})

    @patch.object(PasswordManager, '_get_token', new=lambda x: 'my-fake-token')
    @patch('requests.post')
    def test_request_to_get_password_returns_error(self, mock_post):
        mock_response = Mock()
        mock_response.json.return_value = {'status': 'error', 'password': ''}
        mock_post.return_value = mock_response
        pm = PasswordManager(key='tests/fixtures/privkey.pem')

        passwords = pm.get_passwords('folder-name', 'password-name')

        self.assertDictEqual(passwords, {})

    @patch.object(PasswordManager, '_get_token', new=lambda x: 'my-fake-token')
    @patch('requests.post')
    def test_gets_passwords_correctly(self, mock_post):
        pm = PasswordManager(key='tests/fixtures/privkey.pem')

        def encrypt_text(text):
            public_key = pm._rsa_verifier.public_key()
            encrypted_key = public_key.encrypt(text.encode('ascii'), padding.PKCS1v15())
            return base64.b64encode(bytes(encrypted_key))

        mock_response = Mock()
        mock_response.json.return_value = {
            'status': 'success',
            'password': {
                'url': encrypt_text('a'),
                'login': encrypt_text('b'),
                'passwd': encrypt_text('c'),
                'description': encrypt_text('d'),
                'vault': False
            }
        }
        mock_post.return_value = mock_response

        passwords = pm.get_passwords('folder-name', 'password-name')

        self.assertEqual(passwords['password-name']['url'], 'a')
        self.assertEqual(passwords['password-name']['login'], 'b')
        self.assertEqual(passwords['password-name']['password'], 'c')
        self.assertEqual(passwords['password-name']['description'], 'd')

    @patch.object(PasswordManager, '_get_token', new=lambda x: 'my-fake-token')
    @patch('requests.post')
    def test_gets_passwords_correctly_vault_true(self, mock_post):
        pm = PasswordManager(key='tests/fixtures/privkey.pem')

        def b64_encode(text):
            return base64.b64encode(bytes(text, 'utf-8')).decode()

        mock_response = Mock()
        mock_response.json.return_value = {
            'status': 'success',
            'password': {
                'url': b64_encode('a'),
                'login': b64_encode('b'),
                'passwd': b64_encode('c'),
                'description': b64_encode('d'),
                'vault': True
            }
        }
        mock_post.return_value = mock_response

        passwords = pm.get_passwords('folder-name', 'password-name')
        self.assertEqual(passwords['password-name']['url'], 'a')
        self.assertEqual(passwords['password-name']['login'], 'b')
        self.assertEqual(passwords['password-name']['password'], 'c')
        self.assertEqual(passwords['password-name']['description'], 'd')

    @patch.object(PasswordManager, '_get_token', new=lambda x: 'my-fake-token')
    @patch('requests.post')
    def test_gets_passwords_correctly_vault_missing(self, mock_post):
        pm = PasswordManager(key='tests/fixtures/privkey.pem')

        def encrypt_text(text):
            public_key = pm._rsa_verifier.public_key()
            encrypted_key = public_key.encrypt(text.encode('ascii'), padding.PKCS1v15())
            return base64.b64encode(bytes(encrypted_key))

        mock_response = Mock()
        mock_response.json.return_value = {
            'status': 'success',
            'password': {
                'url': encrypt_text('a'),
                'login': encrypt_text('b'),
                'passwd': encrypt_text('c'),
                'description': encrypt_text('d'),
            }
        }
        mock_post.return_value = mock_response

        passwords = pm.get_passwords('folder-name', 'password-name')
        self.assertEqual(passwords['password-name']['url'], 'a')
        self.assertEqual(passwords['password-name']['login'], 'b')
        self.assertEqual(passwords['password-name']['password'], 'c')
        self.assertEqual(passwords['password-name']['description'], 'd')
